""" contexts_to_pagePacker.py

    This Python script makes use of Applescript to export the context lists out
    of OmniFocus, and then creates a PagePacker document containing those
    context lists.


    ABOUT THIS SCRIPT
    -----------------

    Unfortunately, there is no way to script the "Save as PDF" option in the
    Print dialog box.  This means that it is not possible, using Applescript
    alone, to automatically create a PagePacker document out of a series of
    printed context lists.  This script uses a different approach; it uses
    Applescript to extract the raw text of the context lists out of OmniFocus,
    and then converts that text into a format suitable for importing into
    PagePacker.

    Eventually, we will use ReportLab to generate a PDF version of each context
    list.  For now, we simply import the context lists as semi-formatted text.
    The result is ugly, but at least it works.


    HOW TO USE THIS SCRIPT
    ----------------------

    To run this script, place the "contexts_to_pagePacker.py" file (ie, this
    file) into a suitable directory.  In this example, we will place it into a
    directory named "OmniFocus-to-PagePacker" on the Desktop.

    Now, open the Terminal, and type:

        cd ~/Desktop/OmniFocus-to-PagePacker
        python contexts_to_pagePacker.py

    If you've put the script into a different folder, just change the first
    line to whatever folder name and location you've used.

    If you want to use a different paper size, just edit the PAPER_SIZE setting
    below.

    The script extracts your various context lists and places them into the
    PagePacker document.  You can also add other PDF documents to PagePacker;
    if you place a file named "page-1.pdf" into the same directory as this
    script, it will be placed onto page one of the PagePacker document.  You
    can have similar PDF files for pages 2 through 8.
    
    If you do have one or more predefined PDF files, the context lists will be
    placed on whatever pages are left over.  This lets you do things like add a
    cover page to your document, add an address list to the end, or whatever
    you like.

    Obviously, there have to be enough pages for your various context lists.
    If you have too many contexts, only the first ones will appear in the
    document, and the rest will be ignored.


    RUNNING THE SCRIPT VIA QUICKSILVER
    ----------------------------------

    If you want to run this script automatically, here's one way to do so using
    Quicksilver.  Create a Terminal settings file which automatically runs this
    script, and use Quicksilver to access it.  To do this, create a new
    terminal setting script and set the ExecutionString entry to:

        cd ~/Desktop/OmniFocus-to-PagePacker; python contexts_to_pagePacker.py

    Store your settings file in ~/Library/Application Suport/Terminal.  Then
    add this directory as a custom catalogue entry to QuickSilver, and you can
    immediatley run the script using QuickSilver.

    There may be other ways to do this, but this scheme works well for me.


    FUTURE ENHANCEMENTS
    -------------------

    This script was written quickly to prove that it can be done.  Future
    enhancements include:

        * Displaying the notes associated with a task.

        * Using the "ReportLab" PDF generation toolkit to create a
          nicely-formatted context list, rather than using plain text.

    Other suggestions for ways to improve this script are most welcome.


    ABOUT THE AUTHOR
    ----------------

    My name is Erik Westra, and I'm a software developer based in Rotorua, New
    Zealand.  My email address is ewestra at gmail dot com.

    Please feel free to use this in any way you like.  I hope you find it
    useful!
"""
import os
import os.path

#############################################################################

# Change the following constants to configure the way this script runs:

PAPER_SIZE = "A4" # Use "A4" or "letter"

#############################################################################

# The following string contains the AppleScript code needed to extract the
# context lists out of OmniFocus.
#
# This script creates a series of text files, named "context-1.txt",
# "context-2.txt", etc.  Each file consists of the name of the context as the
# first line, followed by the tasks in that context.  Each task consists of the
# task name, a tab, and the name of the associated project.
#
# Note that we use a special string, "$$CUR_DIR$$", to represent the directory
# to save the context files into.  This string is dynamically substituted
# before the script is run.

GET_CONTEXTS_SCRIPT = """
tell application "OmniFocus"
    set lstContexts to every context of first document
    my ListContexts(lstContexts, "")
end tell

on ListContexts(lstContexts, strParentPath)
    -- Save the contexts into a bunch of separate text files named
    -- "context-1.txt", "context-2.txt", etc.
    using terms from application "OmniFocus"
        set contextNum to 1
        repeat with oContext in lstContexts
            --CONTEXT NAME
            set strPath to name of oContext
            if length of strParentPath > 0 then
                set strPath to strParentPath & " > " & strPath
            end if
            set strList to strPath & return

            --ACTIONS
            set lstTasks to every remaining task of oContext
            if lstTasks = {} then
            else
                set strList to ListTasks(lstTasks, strList)
            end if

            -- SUB-CONTEXTS
            set lstSubContexts to every context of oContext
            if lstSubContexts = {} then
            else
                set strList to ListContexts(lstSubContexts, strPath, strList)
            end if
            my saveToFile("context-" & contextNum & ".txt", strList)
            set contextNum to contextNum + 1
        end repeat
    end using terms from
end ListContexts

on ListTasks(lstTasks, strList)
    using terms from application "OmniFocus"
        repeat with oTask in lstTasks
            set oProject to containing project of oTask
            if oProject is missing value then
                set projectName to ""
            else
                set projectName to name of oProject
            end if
            set strList to strList & name of oTask & tab & projectName & return
        end repeat
        return strList
    end using terms from
end ListTasks

on saveToFile(fName, txt)
    -- Save the given text to the given file.
    set curDir to "$$CUR_DIR$$"
    set fName to curDir & ":" & fName
    set f to (open for access file fName with write permission)
    write txt to f starting at eof
    close access f
end saveToFile
"""

#############################################################################

def deleteTemporaryFiles():
    """ Delete our temporary files, if any.
    """
    for contextNum in range(1, 9):
        fName = "context-" + str(contextNum) + ".txt"
        if os.path.exists(fName):
            os.remove(fName)

#############################################################################

def runApplescript(src):
    """ Run the given Applescript.

        Upon completion, we return True iff the applescript was run
        successfully.

        Note that we suppress the messages written to stdout, as this looks
        silly when running an Applescript using osascript.
    """
    f = os.popen("osascript<<END\n" + src + "\nEND", "r")
    while True:
        response = f.read()
        if len(response) == 0: break
    errCode = f.close()
    if errCode != None:
        print "Error running Applescript: " + repr(errCode)
        return False
    else:
        return True

#############################################################################

def main():
    """ Our main program.
    """
    # Get our current directory, in Appescript format.

    curDir = os.getcwd()
    curDir = curDir[1:].replace("/", ":")

    # Extract the context lists into our temporary files.

    deleteTemporaryFiles()

    src = GET_CONTEXTS_SCRIPT.replace("$$CUR_DIR$$", curDir)
    if not runApplescript(src): return

    # Massage the temporary files to make them readable.  Eventually, this will
    # create PDF documents based on the exported text.  For now, we simply
    # format the text.

    contextNum = 1
    while True:
        contextFile = "context-" + str(contextNum) + ".txt"
        if not os.path.exists(contextFile): break

        f = file(contextFile, "rU")
        context = f.readline().strip()
        items = []
        while True:
            line = f.readline()
            if len(line) == 0: break

            line = line.replace("\x00", "").strip() # Fix weird text format.

            if "\t" in line:
                task,project = line.split("\t")
            else: # Should never happen?
                task    = line
                project = ""

            items.append({'task'    : task,
                          'project' : project})
        f.close()

        f = file(contextFile, "w")
        f.write(context + "\n")
        for item in items:
            if len(item['project']) > 0:
                f.write(" * " + item['task'] + " [" + item['project'] + "]\n")
            else:
                f.write(" * " + item['task'] + "\n")
        f.close()

        contextNum = contextNum + 1

    # Build the Applescript used to import the temporary files into PagePacker.

    src = []
    src.append('tell application "PagePacker"')
    src.append('    set page size to ' + PAPER_SIZE)
    src.append('    make new document')
    src.append('    tell document 1')

    contextNum = 1
    for pageNum in range(1, 9):

        # If the user supplied their own PDF file for this page, use it.
        # Otherwise, place the next available context list on this page.

        pdfFile     = "page-"    + str(pageNum)    + ".pdf"
        contextFile = "context-" + str(contextNum) + ".txt"

        if os.path.exists(pdfFile):
            src.append('        set file source of page ' + str(pageNum) +
                              ' to "' + curDir + ":" + pdfFile + '"')
        elif os.path.exists(contextFile):
            src.append('        set file source of page ' + str(pageNum) +
                              ' to "' + curDir + ":" + contextFile + '"')
            contextNum = contextNum + 1

    src.append('    end tell')
    src.append('end tell')

    if not runApplescript("\n".join(src)): return

    # Finally, delete our temporary files so we don't clutter up the directory.

    deleteTemporaryFiles()

#############################################################################

if __name__ == "__main__":
    main()

